#include "RTEngine.h"
#include "Image24.h"
#include "Primitive.h"
#include "PrimSphere.h"
#include "Light.h"
#include <vector>
#include <math.h>

using namespace std;

RTEngine::RTEngine()
{
  return;
}

RTEngine::~RTEngine()
{
  // Clean primitive list
  {
    vector<Primitive*>::iterator iter;

    for(iter = this->PrimitiveList.begin(); iter != this->PrimitiveList.end(); iter++)
      delete *iter;

    this->PrimitiveList.clear();
  }

  // Clean light list
  {
    vector<Light*>::iterator iter;

    for(iter = this->LightList.begin(); iter != this->LightList.end(); iter++)
      delete *iter;

    this->LightList.clear();
  }
}

void RTEngine::AddSphere(Vector3D Position, double Radius, Material m)
{
  PrimSphere *prim = new PrimSphere(Position, Radius);
  prim->m = m;
  this->PrimitiveList.push_back(prim);
}

void RTEngine::AddLight(Vector3D Position, Vector3D Color)
{
  Light *light = new Light(Position, Color);
  this->LightList.push_back(light);
}

void RTEngine::Render(Image24 *img)
{
  Vector3D CameraPos(0.0, 0.0, -5.0);
  double WX1 = -2.0, WX2 = 2.0, WY1 = 1.5, WY2 = -1.5;

  double DX = (WX2 - WX1) / (double)img->width;
  double DY = (WY2 - WY1) / (double)img->height;

  double SX = WX1, SY = WY1;

  int x, y;
  unsigned int *px = img->data;

  // rendering loop
  for(y = 0; y < img->height; y++)
  {
    SX = WX1;
    for(x = 0; x < img->width; x++)
    {
      Ray ray(CameraPos, Vector3D(SX, SY, 0.0));
      ray.Direction = ray.Direction - ray.Origin;
      ray.Direction.Norm();

      Vector3D color = this->Trace(ray, 0);

      // double RGB -> unsigned int RGB
      int r, g, b;
      r = (int)(color.r * 256.0);
      g = (int)(color.g * 256.0);
      b = (int)(color.b * 256.0);

      r = r > 255 ? 255 : r < 0 ? 0 : r;
      g = g > 255 ? 255 : g < 0 ? 0 : g;
      b = b > 255 ? 255 : b < 0 ? 0 : b;

      *px = (r << 16) | (g << 8) | (b);

      // iter
      SX += DX;
      px++;
    }

    // iter
    SY += DY;
  }
}

// Trace
Vector3D RTEngine::Trace(Ray ray, int refl_depth)
{
  Vector3D color(0.0, 0.0, 0.0);
  double dist = 1000000000.0f;
  Primitive *prim = NULL;

  // find closest intersection
  vector<Primitive*>::iterator iter;
  for(iter = this->PrimitiveList.begin(); iter != this->PrimitiveList.end(); iter++)
  {
    double temp_dist;
    if((*iter)->Intersect(ray, temp_dist) == 0)
      continue;

    if(temp_dist < dist)
    {
      prim = *iter;
      dist = temp_dist;
    }
  }

  if(prim == NULL) 
    return color;


  // shader - distance
  //dist += ray.Origin.z;
  //color = Vector3D(1.0 - dist, 1.0 - dist, 1.0 - dist);

  Vector3D pi = ray.Origin + ray.Direction * dist; // point of intersection

  // material color
  Vector3D prim_color = prim->m.Color;

  // lights
  vector<Light*>::iterator lightiter;
  for(lightiter = this->LightList.begin(); lightiter != this->LightList.end(); lightiter++)
  {
    Vector3D L = (*lightiter)->Position - pi;
    L.Norm();
    Vector3D N = prim->Normal(pi);

    // diffusive
    if(prim->m.Diffusive > 0.0)
    {
      double dot = L.Dot(N);
      if (dot > 0.0l)
      {
        double diff = dot * prim->m.Diffusive;
	color += ((*lightiter)->Color * prim_color) * diff;
      }
    }

    // specular
    if(prim->m.Specular > 0.0)
    {
      Vector3D R = L -  N * L.Dot(N) * 2.0l;
      double dot = ray.Direction.Dot(R);
      if (dot > 0)
      {
        dot = pow(dot, 20.0); // dot *= dot; dot *= dot; dot *= dot; dot *= dot;
        double spec = dot * prim->m.Specular;
        color += (*lightiter)->Color * spec;
      }
    }
  }

  // reflection
  double refl = prim->m.Reflective;
  if(refl > 0.0 && refl_depth < 4)
  {
    Vector3D N = prim->Normal(pi);
    Vector3D R = ray.Direction - N * (ray.Direction.Dot(N) * 2.0);
    Vector3D rcol( 0, 0, 0 );
    double dist;
    Ray tempr(pi + R * 0.0001, R);
    rcol = Trace(tempr, refl_depth + 1);
    color += rcol * refl * prim_color;
  } 

  return color;
}


